package main

import (
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"time"
)

// Object to represent a crafting material in a recipe for output JSON
type CraftingMaterial struct {
	ID       int `json:"id"`
	Quantity int `json:"quantity"`
}

// Object to represent an item that can be sold on the market for output JSON
type MarketItem struct {
	ID                int                `json:"id"`
	Job               string             `json:"job"`
	CraftingMaterials []CraftingMaterial `json:"craftingmaterials,omitempty"`
}

type GarlandToolsCraft struct {
	Job         int `json:"job"`
	Lvl         int `json:"lvl"`
	Ingredients []struct {
		ID     int `json:"id"`
		Amount int `json:"amount"`
	} `json:"ingredients"`
}

type GarlandToolsItem struct {
	Item struct {
		Name        string              `json:"name"`
		Description string              `json:"description"`
		ID          int                 `json:"id"`
		StackSize   int                 `json:"stackSize"`
		Craft       []GarlandToolsCraft `json:"craft"`
	} `json:"item"`
}

var (
	lastGarlandToolsRequest time.Time
	jobMap                  map[int]string
	filterJob               string
)

// All market items are run through this function to filter what should be included
// returns true if item should be included in gilgetter
func filterCriteria(craft *GarlandToolsCraft) bool {
	if jobMap[craft.Job] != filterJob || craft.Lvl > 80 { // Ignore everything but specified job crafts level 80 or below
		return false
	}
	return true
}

func main() {

	filterJob = os.Args[1]

	jobMap = make(map[int]string)

	jobMap[8] = "CRP"
	jobMap[9] = "BSM"
	jobMap[10] = "ARM"
	jobMap[11] = "GSM"
	jobMap[12] = "LTW"
	jobMap[13] = "WVR"
	jobMap[14] = "ALC"
	jobMap[15] = "CUL"

	matchFound := false
	for _, job := range jobMap {
		if filterJob == job {
			matchFound = true
		}
	}
	if !matchFound {
		log.Fatalf("filterJob argument must be one of the following: %v", jobMap)
	}

	// Before anything, we make sure the output-data.json file can be created and opened for writing
	f, err := os.Create(fmt.Sprintf("%s.json", filterJob))
	if err != nil {
		log.Fatalf("failed to create file output-data.json: %v", err)
	}
	defer f.Close()

	// This will hold all the market items and data from universalis in a slice
	marketItems := make([]MarketItem, 0)

	// This will be used for quickly testing if an item is a market item
	marketItemsMap := make(map[int]bool)

	// Get all the market item IDs from universalis
	itemIDs, err := getAllMarketable()
	if err != nil {
		log.Fatalf("failed to get all market items from universalis: %v", err)
	}

	// Loop through every item, make a requset to garland tools db and check item criteria
	// if the item matches what we're searching for, extract data and write it to a file
	for _, itemID := range itemIDs {
		item, err := getGarlandToolsItem(itemID)
		if err != nil {
			log.Printf("failed to get item id '%d' from garland tools: %v", itemID, err)
			continue
		}
		log.Printf("checking '%d' - %s", itemID, item.Item.Name)
		marketItemsMap[itemID] = true

		// Loop through crafts to check criteria, if it matches, extract some data
		for _, craft := range item.Item.Craft {
			if !filterCriteria(&craft) { // Ignore everything except what matches our filter
				continue
			}

			log.Printf("adding filtered recipe '%s'", item.Item.Name)
			marketItem := MarketItem{ID: itemID, CraftingMaterials: make([]CraftingMaterial, 0), Job: jobMap[craft.Job]}

			// Loop through the craft ingredients to extract the data needed
			// This is required because I use the label 'quantity' whereas garland uses 'amount'
			for _, ingredient := range craft.Ingredients {
				craftingMaterial := CraftingMaterial{ID: ingredient.ID, Quantity: ingredient.Amount}
				marketItem.CraftingMaterials = append(marketItem.CraftingMaterials, craftingMaterial)
			}

			marketItems = append(marketItems, marketItem)
		}
	}

	outputItems := make([]MarketItem, 0) // This will hold all all the output
	for _, item := range marketItems {
		found := false
		for _, craftingMaterial := range item.CraftingMaterials {
			if _, ok := marketItemsMap[craftingMaterial.ID]; !ok {
				log.Printf("filtering out item '%d' with non-marketable crafting material '%d' - ", item.ID, craftingMaterial.ID)
				found = true
				break
			}
		}
		if found {
			continue
		}
		outputItems = append(outputItems, item)
	}

	// Marshal to JSON and write to file
	jsonBytes, err := json.Marshal(outputItems)
	if err != nil {
		log.Fatalf("failed to marshal outputItems to JSON: %v", err)
	}

	_, err = f.Write(jsonBytes)
	if err != nil {
		log.Fatalf("failed to write json bytes to file: %v", err)
	}
}

// Gets a slice of all marketable item IDs from Universalis
func getAllMarketable() ([]int, error) {
	resp, err := http.Get("https://universalis.app/api/marketable")
	if err != nil {
		return nil, fmt.Errorf("request to universalis failed: %v", err)
	}

	if resp.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("response from universalis expected status is '%d' but got '%d'", http.StatusOK, resp.StatusCode)

	}

	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return nil, fmt.Errorf("failed to read universalis response body: %v", err)

	}

	marketableItems := make([]int, 0)

	err = json.Unmarshal(body, &marketableItems)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal universalis response body: %v", err)
	}

	return marketableItems, nil
}

// Accepts an item ID and reads from local cache if available, otherwise makes request to garland tools
func getGarlandToolsItem(id int) (*GarlandToolsItem, error) {
	var jsonBytes []byte
	if _, err := os.Stat(fmt.Sprintf("item/%d.json", id)); errors.Is(err, os.ErrNotExist) { // if it doesn't exist, download it and cache it locally

		// Play nice with garland tools and rate limit ourselves
		if time.Now().Before(lastGarlandToolsRequest.Add(time.Millisecond * 250)) {
			time.Sleep(time.Millisecond * 250)
		}
		lastGarlandToolsRequest = time.Now()

		resp, err := http.Get(fmt.Sprintf("http://www.garlandtools.org/db/doc/item/en/3/%d.json", id))
		if err != nil {
			return nil, fmt.Errorf("request to garland tools failed: %v", err)
		}

		if resp.StatusCode != http.StatusOK {
			return nil, fmt.Errorf("response from garland tools expected status is '%d' but got '%d'", http.StatusOK, resp.StatusCode)
		}

		jsonBytes, err = ioutil.ReadAll(resp.Body)
		if err != nil {
			return nil, fmt.Errorf("failed to read garland tools response body: %v", err)
		}

		f, err := os.Create(fmt.Sprintf("item/%d.json", id))
		if err != nil {
			return nil, fmt.Errorf("failed to create file item/%d.json: %v", id, err)
		}
		defer f.Close()

		_, err = f.Write(jsonBytes)
		if err != nil {
			return nil, fmt.Errorf("failed to write json bytes to local cache file: %v", err)
		}
	} else { // read from local cache
		f, err := os.Open(fmt.Sprintf("item/%d.json", id))
		if err != nil {
			return nil, fmt.Errorf("failed to open file 'item/%d.json': %v", id, err)
		}
		jsonBytes, err = ioutil.ReadAll(f)
		if err != nil {
			return nil, fmt.Errorf("failed to read file 'item/%d.json': %v", id, err)
		}
	}

	garlandToolsItem := &GarlandToolsItem{}
	err := json.Unmarshal(jsonBytes, garlandToolsItem)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal garland tools response body: %v", err)
	}

	return garlandToolsItem, nil
}
